#!/usr/bin/perl
# ^^^ You may need to change this to the location of your perl interpreter
# RemoteBox v0.2 (c) 2010 Ian Chapman. Licenced under the terms of the GPL
use strict;
use warnings;
use SOAP::Lite;
use FindBin qw($Bin);
use lib "$Bin/share/remotebox";
use vboxService qw($endpoint);
require 'vboxserializers.pl';
require 'rbox_enums.pl';
require 'rbox_gui.pl';
$|=1;

our (%gui, %signal, %sensgrp, %EAudioDriverType, %EAudioControllerType, %ENetworkAdapterType,
     %ENetworkAttachmentType, %ETruth, %EAbled);
my (%hostspec, %osfamily, %osversion, $EIMachineM);
my %prefs = (RDPCLIENT => 'rdesktop %h:%p -T "RemoteBox - %n"');
$endpoint = 'http://localhost:18083';

&set_status_msg("Welcome to $gui{appname} $gui{appver}.");
&get_prefs();
Gtk2->main;

sub get_prefs() {
    if (defined($ENV{HOME})) {
        if (open(PREFS, "$ENV{HOME}/.remotebox.conf")) {
            my @contents = <PREFS>;
            chomp(@contents);
            close(PREFS);
            foreach (@contents) {
                if ($_ =~ m/^URL=(.*)$/) { $prefs{URL}{$1} = 'URL'; }
                elsif ($_ =~ m/^USER=(.*)$/) { $prefs{USER}{$1} = 'USER'; }
                elsif ($_ =~ m/^(.*)=(.*)$/) { $prefs{$1} = $2; }
            }
        }
    }
}

sub save_prefs() {
    if (defined($ENV{HOME})) {
        if (open(PREFS, ">$ENV{HOME}/.remotebox.conf")) {
            foreach my $key (keys %prefs) {
                if ($key eq 'URL') {
                    foreach (keys %{$prefs{URL}}) { print PREFS "URL=$_\n"; }
                }
                elsif ($key eq 'USER') {
                    foreach (keys %{$prefs{USER}}) { print PREFS "USER=$_\n"; }
                }
                else { print PREFS "$key=$prefs{$key}\n"; }
            }
            close(PREFS);
        }
        else { warn "Unable to save preferences: $ENV{HOME}/.remotebox.conf\n"; }
    }
}

sub quit_remotebox() {
    &virtualbox_logoff();
    exit; # Gtk2->main_quit is deprecated
}

sub virtualbox_logon() {
    my ($url, $user, $password) = @_;
    &virtualbox_logoff(); # Ensure we disconnect from an existing connection
    $endpoint = $url;
    eval { $gui{websn} = vboxService->IWebsessionManager_logon($user, $password); };
    return $@;
}

sub virtualbox_logoff() {
    if ($gui{websn}) {
        eval{vboxService->IWebsessionManager_logoff($gui{websn})};
        $gui{websn} = undef;
    }
}

sub reset_guest() {
    my ($widget, $treeview) = @_;
    my $gname = &get_selected($treeview, 0);
    my $guuid = &get_selected($treeview, 5);
    vboxService->IConsole_reset(&get_IConsole($guuid));
    &set_status_msg("Sent request to reset $gname");
}

sub stop_guest_poweroff() {
    my $gname = &get_selected($gui{treeviewGuest}, 0);
    my $guuid = &get_selected($gui{treeviewGuest}, 5);
    vboxService->IConsole_powerDown(&get_IConsole($guuid));
    &get_guest_list();
    &set_status_msg("Sent request to power off $gname. Refresh may be needed to see status changes.");
}

sub stop_guest_acpi() {
    my $gname = &get_selected($gui{treeviewGuest}, 0);
    my $guuid = &get_selected($gui{treeviewGuest}, 5);
    vboxService->IConsole_powerButton(&get_IConsole($guuid));
    &get_guest_list();
    &set_status_msg("Sent ACPI request to shutdown $gname. Refresh may be needed to see status changes.");
}

sub stop_guest_savestate() {
    &busy_window($gui{windowMain});
    my $gname = &get_selected($gui{treeviewGuest}, 0);
    my $guuid = &get_selected($gui{treeviewGuest}, 5);
    my $IProgress = vboxService->IConsole_saveState(&get_IConsole($guuid));
    &set_status_msg("Saving guest state of $gname...");
    vboxService->IProgress_waitForCompletion($IProgress, -1);
    &get_guest_list();
    &set_status_msg("Saved the state of $gname. Refresh may be needed to see status changes.");
    &unbusy_window($gui{windowMain});
}

sub start_guest() {
    my ($widget, $treeview) = @_;
    &busy_window($gui{windowMain});
    my $gname = &get_selected($treeview, 0);
    my $guuid = &get_selected($treeview, 5);
    my $ISession = vboxService->IWebsessionManager_getSessionObject($gui{websn});
    my $IProgress = vboxService->IVirtualBox_openRemoteSession($gui{websn}, $ISession, $guuid, 'vrdp', "");
    if ($IProgress) {
        print "$IProgress s\n";
        &set_status_msg("Attempting to start $gname...");
        vboxService->IProgress_waitForCompletion($IProgress, -1);
        sleep 1; # Wait a small amount for guest to change to running state
        vboxService->ISession_close($ISession);
        &get_guest_list();
        &set_status_msg("Request sent to start $gname. Refresh may be needed to see status changes.");
    }
    else {
        &show_error_dialog('Session in Use',
                           'This guest already has an open session associated with it, either from RemoteBox or another' .
                           ' program. Please close that session and try again.');
        vboxService->ISession_close($ISession);
    }
    &unbusy_window($gui{windowMain});
}

sub delete_guest() {
    my ($widget, $treeview) = @_;
    my ($gname, $gos, $IMachine, $status, $gosid) = &get_selected($treeview);
    &busy_window($gui{windowMain});
    my $response = $gui{dialogDeleteGuest}->run;
    $gui{dialogDeleteGuest}->hide;

    if ($response eq 'ok') {
        &set_status_msg("Deleting $gname...");
        my ($ISession, $IMachineM) = &get_mutable_session($IMachine);
        my $machineid = vboxService->IMachine_getId($IMachineM);
        my @IMediumAttachment = vboxService->IMachine_getMediumAttachments($IMachineM);

        foreach my $attach (@IMediumAttachment) {
            vboxService->IMachine_detachDevice($IMachineM, $$attach{controller}, $$attach{port}, $$attach{device});
        }

        vboxService->IMachine_saveSettings($IMachineM);
        vboxService->ISession_close($ISession);
        vboxService->IVirtualBox_unregisterMachine($gui{websn}, $machineid);
        vboxService->IMachine_deleteSettings($IMachine);
        &get_guest_list();
        &set_status_msg("Deleted $gname...");
    }

    &unbusy_window($gui{windowMain});
}

sub get_mutable_session() {
    my ($IMachine) = @_;
    my $ISession = vboxService->IWebsessionManager_getSessionObject($gui{websn});
    vboxService->IVirtualBox_openSession($gui{websn}, $ISession, vboxService->IMachine_getId($IMachine));
    my $IMachineM = vboxService->ISession_getMachine($ISession);
    return $ISession, $IMachineM;
}

sub recurse_snapshot() {
    my ($treestore, $ISnapshot, $iter) = @_;
    my $citer = $treestore->append($iter);
    my $snapname = vboxService->ISnapshot_getName($ISnapshot);
    my $date = scalar(localtime((vboxService->ISnapshot_getTimeStamp($ISnapshot))/1000)); # VBox returns msecs so / 1000
    $treestore->set($citer, (0 => $snapname, 1 => $date));
    my @snapshots = vboxService->ISnapshot_getChildren($ISnapshot);
    if (@snapshots > 0) { &recurse_snapshot($treestore, $_, $citer) foreach (@snapshots); }
}

sub get_snapshots() {
    my ($widget, $treestore) = @_;
    &busy_window($gui{windowMain});
    &clear_liststore($treestore);
    my $IMachine =  &get_selected($gui{treeviewGuest}, 2);
    &set_status_msg("Retrieving snapshot information from $endpoint...");

    if (vboxService->IMachine_getSnapshotCount($IMachine) > 0) {
        my $ISnapshot = vboxService->IMachine_getSnapshot($IMachine, undef); # get first snapshot
        &recurse_snapshot($treestore, $ISnapshot, undef);
        $gui{treeviewSnapshots}->expand_all();
    }
    else {
        my $iter = $treestore->append(undef);
        $treestore->set($iter, (0 => 'No snapshots available'));
    }
    &set_status_msg('Retrieved snapshot information.');
    &unbusy_window($gui{windowMain});
}

# Returns the IConsole
sub get_IConsole() {
    my ($guuid) = @_;
    my $ISession = vboxService->IWebsessionManager_getSessionObject($gui{websn});
    vboxService->IVirtualBox_openExistingSession($gui{websn}, $ISession, $guuid);
    my $IConsole = vboxService->ISession_getConsole($ISession);
    vboxService->ISession_close($ISession);
    return $IConsole;
}

# Determines the next free port on a controller
sub get_free_deviceport() {
    my ($IMachine, $IStorCtr) = @_;
    my @free = (-1, -1);
    my @used;
    my $ctrname = vboxService->IStorageController_getName($IStorCtr);
    my @IMediumAttachment = vboxService->IMachine_getMediumAttachmentsOfController($IMachine, $ctrname);
    my $hiport = (vboxService->IStorageController_getMaxPortCount($IStorCtr)) - 1;
    my $hidev = (vboxService->IStorageController_getMaxDevicesPerPortCount($IStorCtr)) - 1;
    # Get a list of all used ports
    foreach my $attach (@IMediumAttachment) { $used[$$attach{device}][$$attach{port}] = $attach; }

    foreach my $dev (0..$hidev) {
        last if ($free[0] != -1);

        foreach my $port (0..$hiport) {
            next if ($used[$dev][$port]);
            $free[0] = $dev;
            $free[1] = $port;
            last;
        }
    }

    return @free;
}

sub get_details() {
    my ($widget, $liststore) = @_;
    &busy_window($gui{windowMain});
    &set_status_msg("Retrieving guest details...");
    &clear_liststore($liststore);
    my $IMachine =  &get_selected($gui{treeviewGuest}, 2);
    my $IVRDPServer = vboxService->IMachine_getVRDPServer($IMachine);
    my @IStorageController = vboxService->IMachine_getStorageControllers($IMachine);
    my $IAudioAdapter = vboxService->IMachine_getAudioAdapter($IMachine);
    my $IUSBController = vboxService->IMachine_getUSBController($IMachine);

    &append_liststore($liststore, (0 => 'GENERAL', 2 => 'double'));
    &append_liststore($liststore, (0 => 'Name:', 1 => vboxService->IMachine_getName($IMachine)));
    &append_liststore($liststore, (0 => 'OS Type:', 1 => vboxService->IMachine_getOSTypeId($IMachine)));
    &append_liststore($liststore, (0 => 'SYSTEM', 2 => 'double'));
    &append_liststore($liststore, (0 => 'Firmware:', 1 => vboxService->IMachine_getFirmwareType($IMachine)));
    &append_liststore($liststore, (0 => 'Base Memory:', 1 => vboxService->IMachine_getMemorySize($IMachine) . ' MiB'));
    &append_liststore($liststore, (0 => 'Processor(s):', 1 => vboxService->IMachine_getCPUCount($IMachine)));

    foreach (1..4) {
        my $bootdev = vboxService->IMachine_getBootOrder($IMachine,$_);
        &append_liststore($liststore, (0 => "Boot Device $_:", 1 => $bootdev)) if ($bootdev ne 'Null');
    }

    &append_liststore($liststore, (0 => 'VT-x/AMV-V:', 1 => $EAbled{vboxService->IMachine_getHWVirtExProperty($IMachine, 'Enabled')}));
    &append_liststore($liststore, (0 => 'VT-x VPID:', 1 => $EAbled{vboxService->IMachine_getHWVirtExProperty($IMachine, 'VPID')}));
    &append_liststore($liststore, (0 => 'Nested Paging:', 1 => $EAbled{vboxService->IMachine_getHWVirtExProperty($IMachine, 'NestedPaging')}));
    &append_liststore($liststore, (0 => 'DISPLAY', 2 => 'double'));
    &append_liststore($liststore, (0 => 'Video Memory:', 1 => vboxService->IMachine_getVRAMSize($IMachine) . ' MiB'));
    &append_liststore($liststore, (0 => '3D Acceleration:', 1 => $EAbled{vboxService->IMachine_getAccelerate3DEnabled($IMachine)}));
    &append_liststore($liststore, (0 => '2D Acceleration:', 1 => $EAbled{vboxService->IMachine_getAccelerate2DVideoEnabled($IMachine)}));

    if (vboxService->IVRDPServer_getEnabled($IVRDPServer) eq 'true') {
        &append_liststore($liststore, (0 => 'Remote Display Server:', 1 => 'Enabled'));
        &append_liststore($liststore, (0 => 'Remote Display Port(s):', 1 => vboxService->IVRDPServer_getPorts($IVRDPServer)));
    }
    else { &append_liststore($liststore, (0 => 'Remote Display Server:', 1 => 'Disabled')); }

    &append_liststore($liststore, (0 => 'STORAGE', 2 => 'double'));

    foreach my $controller (@IStorageController) {
        my $controllername = vboxService->IStorageController_getName($controller);
        &append_liststore($liststore, (0 => 'Controller:', 1 => $controllername));
        my @IMediumAttachment = vboxService->IMachine_getMediumAttachmentsOfController($IMachine, $controllername);
        foreach my $attachment (@IMediumAttachment) {
            if ($$attachment{medium}) {
                &append_liststore($liststore, (0 => "  Port $$attachment{port}:",
                                               1 => "$$attachment{type}: " . vboxService->IMedium_getName($$attachment{medium})));
            }
        }
    }

    &append_liststore($liststore, (0 => 'AUDIO', 2 => 'double'));

    if (vboxService->IAudioAdapter_getEnabled($IAudioAdapter) eq 'true') {
        &append_liststore($liststore, (0 => 'Host Driver:', 1 => $EAudioDriverType{vboxService->IAudioAdapter_getAudioDriver($IAudioAdapter)}{desc}));
        &append_liststore($liststore, (0 => 'Controller:', 1 => $EAudioControllerType{vboxService->IAudioAdapter_getAudioController($IAudioAdapter)}));
    }
    else { &append_liststore($liststore, (0 => 'Audio Disabled')); }

    &append_liststore($liststore, (0 => 'NETWORK', 2 => 'double'));

    foreach (0..($hostspec{maxnet}-1)) {
        my $INetworkAdapter = vboxService->IMachine_getNetworkAdapter($IMachine,$_);
        if (vboxService->INetworkAdapter_getEnabled($INetworkAdapter) eq 'true') {
            &append_liststore($liststore, (0 => "Adapter $_:",
                                           1 => $ENetworkAdapterType{vboxService->INetworkAdapter_getAdapterType($INetworkAdapter)} . ', ' .
                                                                     vboxService->INetworkAdapter_getAttachmentType($INetworkAdapter)));
        }
    }

    &append_liststore($liststore, (0 => 'SERIAL PORTS', 2 => 'double'));

    foreach (0..($hostspec{maxser}-1)) {
        my $ISerialPort = vboxService->IMachine_getSerialPort($IMachine,$_);
        if (vboxService->ISerialPort_getEnabled($ISerialPort) eq 'true') {
            &append_liststore($liststore, (0 => "Port #$_:",
                                           1 => 'Enabled, ' . vboxService->ISerialPort_getHostMode($ISerialPort) . ', ' .
                                                              vboxService->ISerialPort_getPath($ISerialPort)));
        }
        else { &append_liststore($liststore, (0 => "Port #$_:", 1 => 'Disabled')); }
    }

    &append_liststore($liststore, (0 => 'USB', 2 => 'double'));

    if (vboxService->IUSBController_getEnabled($IUSBController) eq 'true' ) {
        my @IUSBDeviceFilter = vboxService->IUSBController_getDeviceFilters($IUSBController);
        my $active = 0;
        my $inactive = 0;
        foreach my $filter (@IUSBDeviceFilter) {
            if (vboxService->IUSBDeviceFilter_getActive($filter) eq 'true') { $active++; }
            else { $inactive++; }
        }
        &append_liststore($liststore, (0 => 'Active Device Filters:', 1 => $active));
        &append_liststore($liststore, (0 => 'Inactive Device Filters:', 1 => $inactive));
    }
    else { &append_liststore($liststore, (0 => 'Disabled')); }

    my @sf = vboxService->IMachine_getSharedFolders($IMachine);
    &append_liststore($liststore, (0 => 'SHARED FOLDERS', 2 => 'double'));
    &append_liststore($liststore, (0 => 'Shared Folders:',1 => scalar(@sf)));
    &set_status_msg('Retrieved guest details.');
    &unbusy_window($gui{windowMain});
}

sub show_connect_dialog() {
    my ($widget, $dialog) = @_;

    &busy_window($gui{windowMain});
    &clear_liststore($gui{liststoreConnectURL}, $gui{liststoreConnectUser});
    &append_liststore($gui{liststoreConnectURL}, (0 => $_)) foreach (keys (%{$prefs{URL}}));
    &append_liststore($gui{liststoreConnectUser}, (0 => $_)) foreach (keys(%{$prefs{USER}}));
    my $response = $dialog->run;
    $dialog->hide;
    $dialog->get_display->flush;

    if ($response eq 'ok') {
        my $url = $gui{comboboxentryConnectURL}->get_active_text();
        my $user = $gui{comboboxentryConnectUser}->get_active_text();
        $url = $endpoint if (!$url);
        $url = "http://$url" if ($url !~ m/^.+:\/\//);
        $url = "$url:18083" if ($url !~ m/:\d+$/);

        if ($gui{checkbuttonConnectSave}->get_active()) {
            $prefs{URL}{$url} = 'URL' if ($url);
            $prefs{USER}{$user} = 'USER' if ($user);
            &save_prefs();
        }

        my $err = &virtualbox_logon($url, $user, $gui{entryConnectPassword}->get_text());
        if ($err) {
            $err =~ s/at.+$//;
            &show_error_dialog('Failed to Connect',
                               'Your credentials may be incorrect or the vbox' .
                               " web service is not running at the specified URL.\n\n" .
                               "Additional:\n$err");
            &set_status_msg("Unable to connect to $url");
        }
        else {
            my $ver = vboxService->IVirtualBox_getVersion($gui{websn});

            if (!$ver) {
                &show_error_dialog('Failed to Authenticate',
                                   'Login credentials are incorrect or the web service is mis-configured.');
                &set_status_msg("Failed to authenticate to $url");
                &virtualbox_logoff();
            }
            else {

                if ($ver !~ m/^3.2/) {
                    &show_warn_dialog("Unsupported VirtualBox Version ($ver)",
                                      "This version of RemoteBox is not intended for use with the version of VirtualBox" .
                                      " running on the server.\n\nRemoteBox will continue, but due to changes in the" .
                                      " APIs of VirtualBox, you may experience failures or loss of functionality.\n\n" .
                                      "VirtualBox 3.2.x is recommended.");

                }

                if ($ver =~ m/_OSE$/) {
                    &show_warn_dialog("OSE Version Detected ($ver)",
                                      "The OSE version of VirtualBox is not yet officially supported by RemoteBox.\n\n" .
                                      "RemoteBox will continue, but you may experience failures or loss of functionality");
                }

                &populate_hostspec();
                &populate_ostypes();
                &get_guest_list();
                &sensitive_on(@{$sensgrp{connect}});
                &set_status_msg("Logged onto $endpoint running VirtualBox $hostspec{vbver}");
            }
        }
    }

    &unbusy_window($gui{windowMain});
}

sub show_about_dialog() {
    my ($widget, $dialog) = @_;
    $dialog->run;
    $dialog->hide;
}

sub show_serverinfo_dialog() {
    my ($widget, $liststore) = @_;
    &busy_window($gui{windowMain});
    &clear_liststore($liststore);
    &append_liststore($liststore, (0 => 'URL:', 1 => $endpoint));
    &append_liststore($liststore, (0 => 'VirtualBox Version:', 1 => $hostspec{vbver}));
    &append_liststore($liststore, (0 => 'Build Revision:', 1 => $hostspec{buildrev}));
    &append_liststore($liststore, (0 => 'Package Type:',1 => $hostspec{pkgtype}));
    &append_liststore($liststore, (0 => 'Global Settings File:', 1 => $hostspec{settingsfile}));
    &append_liststore($liststore, (0 => 'Machine Folder:', 1 => $hostspec{machinedir}));
    &append_liststore($liststore, (0 => 'Hard Disk Folder:', 1 => $hostspec{hddir}));
    &append_liststore($liststore, (0 => 'Server Logical CPUs:', 1 => $hostspec{maxhostcpuon}));
    &append_liststore($liststore, (0 => 'Server CPU Type:', 1 => $hostspec{cpudesc}));
    &append_liststore($liststore, (0 => 'Server CPU Speed:', 1 => "$hostspec{cpuspeed} Mhz (approx)"));
    &append_liststore($liststore, (0 => 'Server Memory Size:', 1 =>  "$hostspec{memsize} MiB"));
    &append_liststore($liststore, (0 => 'Server OS:', 1 => $hostspec{os}));
    &append_liststore($liststore, (0 => 'Server OS Version:',1 => $hostspec{osver}));
    &append_liststore($liststore, (0 => 'Min Guest RAM:', 1 => "$hostspec{minguestram} MiB"));
    &append_liststore($liststore, (0 => 'Max Guest RAM:', 1 => "$hostspec{maxguestram} MiB"));
    &append_liststore($liststore, (0 => 'Min Guest Video RAM:', 1 => "$hostspec{minguestvram} MiB"));
    &append_liststore($liststore, (0 => 'Max Guest Video RAM:', 1 => "$hostspec{maxguestvram} MiB"));
    &append_liststore($liststore, (0 => 'Max Guest CPUs:', 1 => $hostspec{maxguestcpu}));
    &append_liststore($liststore, (0 => 'Max VDI Size:', 1 => "$hostspec{maxhdsize} MiB"));
    $gui{dialogServerInfo}->run;
    $gui{dialogServerInfo}->hide;
    &unbusy_window($gui{windowMain});
}

sub show_createhd_dialog() {
    my ($widget) = @_;
    &busy_window($gui{dialogVMM});
    $gui{entryCreateHDName}->set_text('NewHardDisk' . int(rand(9999)));
    $gui{spinbuttonCreateHDSize}->set_range(8.00, $hostspec{maxhdsize});
    $gui{spinbuttonCreateHDSize}->set_value(2048.00);
    my $response = $gui{dialogCreateHD}->run;
    $gui{dialogCreateHD}->hide;
    my $diskname = $gui{entryCreateHDName}->get_text();
    $diskname =~ s/\.vdi$//;
    my $disksize = $gui{spinbuttonCreateHDSize}->get_value_as_int();

    if ($diskname and ($response eq 'ok')) {
        if ($gui{checkbuttonCreateHDDynamic}->get_active()) { &create_new_HD($diskname, 'VDI', 'Standard', $disksize); }
        else { &create_new_HD($diskname, 'VDI', 'Fixed', $disksize); }
        &vmm_mediahd_fill();
    }
    &unbusy_window($gui{dialogVMM});
}

sub show_vmm_dialog() {
    my ($widget) = @_;
    &busy_window($gui{windowMain});
    &set_status_msg("Retrieving media information...");
    &vmm_mediahd_fill();
    &vmm_mediadvd_fill();
    &vmm_mediafloppy_fill();
    &set_status_msg("Retrieved media information.");
    $gui{dialogVMM}->run;
    $gui{dialogVMM}->hide;
    &unbusy_window($gui{windowMain});
}

sub vmm_mediahd_fill() {
    &clear_liststore($gui{liststoreVMMHD});
    my @IMedium = vboxService->IVirtualBox_getHardDisks($gui{websn});
    foreach my $hd (@IMedium) {
        my $gnames;
        vboxService->IMedium_refreshState($hd); # Needed otherwise sizes may be 0
        my $name = vboxService->IMedium_getName($hd);
        my $vsize = vboxService->IMedium_getLogicalSize($hd);
        my $asize = int(vboxService->IMedium_getSize($hd) / 1048576);
        my @guuids = vboxService->IMedium_getMachineIds($hd);

        foreach my $id (@guuids) {
            my $IMachine = vboxService->IVirtualBox_getMachine($gui{websn}, $id);
            $gnames .= vboxService->IMachine_getName($IMachine) . ",";
        }

        if ($gnames) {
            chop($gnames); # Remove trailing ,
            if (length($gnames) > 50) { $gnames = substr($gnames, 0, 50) . '... +more'; }
        }
        else { $gnames = '<None>'; }

        &append_liststore($gui{liststoreVMMHD}, (0 => $name,
                                                 1 => $hd,
                                                 2 => $gnames,
                                                 3 => "$asize (MiB)",
                                                 4 => "$vsize (MiB)"));
    }
}

sub vmm_mediadvd_fill() {
    &clear_liststore($gui{liststoreVMMDVD});
    my @IMedium = vboxService->IVirtualBox_getDVDImages($gui{websn});
    foreach my $dvd (@IMedium) {
        my $gnames;
        vboxService->IMedium_refreshState($dvd); # Needed otherwise sizes may be 0
        my $name = vboxService->IMedium_getName($dvd);
        my $size = int(vboxService->IMedium_getSize($dvd) / 1048576);
        my @guuids = vboxService->IMedium_getMachineIds($dvd);

        foreach my $id (@guuids) {
            my $IMachine = vboxService->IVirtualBox_getMachine($gui{websn}, $id);
            $gnames .= vboxService->IMachine_getName($IMachine) . ",";
        }

        if ($gnames) {
            chop($gnames); # Remove trailing ,
            if (length($gnames) > 50) { $gnames = substr($gnames, 0, 50) . '... +more'; }
        }
        else { $gnames = '<None>'; }

        &append_liststore($gui{liststoreVMMDVD}, (0 => $name,
                                                  1 => $dvd,
                                                  2 => $gnames,
                                                  3 => "$size (MiB)"));
    }
}

sub vmm_mediafloppy_fill() {
    &clear_liststore($gui{liststoreVMMFloppy});
    my @IMedium = vboxService->IVirtualBox_getFloppyImages($gui{websn});
    foreach my $floppy (@IMedium) {
        my $gnames;
        vboxService->IMedium_refreshState($floppy); # Needed otherwise sizes may be 0
        my $name = vboxService->IMedium_getName($floppy);
        my $size = vboxService->IMedium_getSize($floppy);
        my @guuids = vboxService->IMedium_getMachineIds($floppy);

        foreach my $id (@guuids) {
            my $IMachine = vboxService->IVirtualBox_getMachine($gui{websn}, $id);
            $gnames .= vboxService->IMachine_getName($IMachine) . ",";
        }

        if ($gnames) {
            chop($gnames); # Remove trailing ,
            if (length($gnames) > 50) { $gnames = substr($gnames, 0, 50) . '... +more'; }
        }
        else { $gnames = '<None>'; }

        &append_liststore($gui{liststoreVMMFloppy}, (0 => $name,
                                                     1 => $floppy,
                                                     2 => $gnames,
                                                     3 => "$size (Bytes)"));
    }
}

sub show_prefs_dialog() {
    my ($widget) = @_;
    &busy_window($gui{windowMain});
    $gui{entryPrefsRDPClient}->set_text($prefs{RDPCLIENT});
    my $response = $gui{dialogPrefs}->run;
    $gui{dialogPrefs}->hide;

    if ($response eq 'ok') {
        $prefs{RDPCLIENT} = $gui{entryPrefsRDPClient}->get_text();
        &save_prefs();
    }
    &unbusy_window($gui{windowMain});
}

sub show_new_dialog() {
    my ($widget, $dialog) = @_;
    &busy_window($gui{windowMain});
    my @IMedium = vboxService->IVirtualBox_getHardDisks($gui{websn});
    $gui{entryNewName}->set_text('NewGuest' . int(rand(9999)));
    $gui{notebookNew}->set_current_page(0);
    $gui{spinbuttonNewMemory}->set_range(4, $hostspec{memsize});
    $gui{spinbuttonNewHDSize}->set_range(8.00, $hostspec{maxhdsize});
    $gui{comboboxNewOSFam}->signal_handler_block($signal{fam}); # Block to avoid signal emission when changing
    $gui{comboboxNewOSVer}->signal_handler_block($signal{ver});
    &clear_liststore($gui{liststoreNewOSFam}, $gui{liststoreNewOSVer}, $gui{liststoreNewChooseHD});

    foreach my $fam (sort(keys %osfamily)) {
        my $iter = &append_liststore($gui{liststoreNewOSFam}, (0 => "$osfamily{$fam}{description}",
                                                               1 => $fam ));
        $gui{comboboxNewOSFam}->set_active_iter($iter) if ($fam eq 'Windows');
    }

    if (@IMedium > 0) {
        &append_liststore($gui{liststoreNewChooseHD}, (0 => vboxService->IMedium_getName($_), 1=> $_)) foreach (@IMedium);
        $gui{comboboxNewChooseHD}->set_active(0);
        &sensitive_on($gui{radiobuttonNewExistingHD});
    }
    else { &sensitive_off($gui{radiobuttonNewExistingHD}) };

    $gui{comboboxNewOSFam}->signal_handler_unblock($signal{fam});
    $gui{comboboxNewOSVer}->signal_handler_unblock($signal{ver});
    $gui{comboboxNewOSFam}->signal_emit('changed');    # Force update of other fields based on OS
    $gui{comboboxNewOSVer}->signal_emit('changed'); # Force update of other fields based on OS
    my $response = $dialog->run;
    $dialog->hide;

    if ($response eq 'ok') {
        &busy_window($dialog);
        my $buttongrp = $gui{radiobuttonNewNewHD}->get_group();
        my %guest = (name   => $gui{entryNewName}->get_text(),
                     fam    => &get_selected($gui{comboboxNewOSFam}, 1),
                     ver    => &get_selected($gui{comboboxNewOSVer}, 1),
                     mem    => $gui{spinbuttonNewMemory}->get_value_as_int(),
                     hdsize => $gui{spinbuttonNewHDSize}->get_value_as_int());

        $guest{name} = 'NewGuest' . int(rand(9999)) if (!$guest{name});
        my ($IMachine, $dvdctrname, $hdctrname) = &create_new_guest($guest{name},$guest{fam}, $guest{ver}, $guest{mem});

        if ($IMachine) {
            my ($ISession, $IMachineM) = &get_mutable_session($IMachine);
            my $IMediumHD;

            if ($$buttongrp[0]->get_active() == 1){ $IMediumHD = &get_selected($gui{comboboxNewChooseHD}, 1); }
            else {
                my $hdimgtype = 'Fixed';
                $hdimgtype = 'Standard' if ($gui{checkbuttonNewDynamic}->get_active());
                $IMediumHD = &create_new_HD($guest{name}, 'VDI', $hdimgtype, $guest{hdsize});
            }

            my $IStorCtrHD = vboxService->IMachine_getStorageControllerByName($IMachineM, $hdctrname);
            &create_sas_sata_port($IStorCtrHD);
            my ($hddev, $hdport) = &get_free_deviceport($IMachineM, $IStorCtrHD);
            vboxService->IMachine_attachDevice($IMachineM, $hdctrname, $hdport, $hddev, 'HardDisk', vboxService->IMedium_getId($IMediumHD));

            # Attach Empty CD/DVD Device
            my $IStorCtrDVD = vboxService->IMachine_getStorageControllerByName($IMachineM, $dvdctrname);
            &create_sas_sata_port($IStorCtrDVD);
            my ($dvddev, $dvdport) = &get_free_deviceport($IMachineM, $IStorCtrDVD);
            vboxService->IMachine_attachDevice($IMachineM, $dvdctrname, $dvdport, $dvddev, 'DVD', 'UUID 00000000-0000-0000-0000-000000000000');
            vboxService->IMachine_saveSettings($IMachineM);
            vboxService->ISession_close($ISession);
        }
        else {
            &show_error_dialog('Failed to Create Guest.',
                               'This may because a guest with that name already exists. For further details' .
                               ' please consult the web logs.');
            &set_status_msg("Failed to create the new guest $guest{name}");
        }
        &unbusy_window($dialog);
        &get_guest_list();
    }
    &unbusy_window($gui{windowMain});
}

sub show_edit_dialog() {
    my ($widget, $treeview) = @_;
    my ($gname, $gos, $IMachine, $status, $gosid) = &get_selected($treeview);
    &busy_window($gui{windowMain});
    &set_status_msg("Retrieving guest settings for $gname...");
    $gui{dialogEdit}->set_title("Edit Settings - $gname");
    my ($ISession, $IMachineM) = &get_mutable_session($IMachine);

    if ($IMachineM) {
        $EIMachineM = $IMachineM;
        $gui{comboboxEditGenOSFam}->signal_handler_block($signal{famedit});
        $gui{comboboxEditGenOSVer}->signal_handler_block($signal{veredit});
        $gui{comboboxEditAudioDriver}->signal_handler_block($signal{audiodrv});
        $gui{comboboxEditAudioCtr}->signal_handler_block($signal{audioctr});
        $gui{comboboxEditNetType1}->signal_handler_block($signal{netadaptype1});
        $gui{comboboxEditNetType2}->signal_handler_block($signal{netadaptype2});
        $gui{comboboxEditNetType3}->signal_handler_block($signal{netadaptype3});
        $gui{comboboxEditNetType4}->signal_handler_block($signal{netadaptype4});
        $gui{comboboxEditNetAttach1}->signal_handler_block($signal{netattach1});
        $gui{comboboxEditNetAttach2}->signal_handler_block($signal{netattach2});
        $gui{comboboxEditNetAttach3}->signal_handler_block($signal{netattach3});
        $gui{comboboxEditNetAttach4}->signal_handler_block($signal{netattach4});
        $gui{comboboxEditNetName1}->signal_handler_block($signal{netname1});
        $gui{comboboxEditNetName2}->signal_handler_block($signal{netname2});
        $gui{comboboxEditNetName3}->signal_handler_block($signal{netname3});
        $gui{comboboxEditNetName4}->signal_handler_block($signal{netname4});

        &clear_liststore($gui{liststoreEditGenOSFam}, $gui{liststoreEditGenOSVer}, $gui{liststoreEditAudioDriver},
                        $gui{liststoreEditAudioCtr}, $gui{liststoreEditNetAdapterType1}, $gui{liststoreEditNetAdapterType2},
                        $gui{liststoreEditNetAdapterType3}, $gui{liststoreEditNetAdapterType4}, $gui{liststoreEditNetAttachTo1},
                        $gui{liststoreEditNetAttachTo2}, $gui{liststoreEditNetAttachTo3}, $gui{liststoreEditNetAttachTo4},
                        $gui{liststoreEditNetName1}, $gui{liststoreEditNetName2}, $gui{liststoreEditNetName3},
                        $gui{liststoreEditNetName4}, $gui{liststoreEditSysBoot});
        # These widgets must start insensitive
        &sensitive_off($gui{buttonEditSysBootUp}, $gui{buttonEditSysBootDown});
        &setup_edit_dialog_general($IMachineM, $treeview);
        &setup_edit_dialog_system($IMachineM);
        &setup_edit_dialog_display($IMachineM);
        &setup_edit_dialog_storage($IMachineM);
        &setup_edit_dialog_audio($IMachineM);
        &setup_edit_dialog_network($IMachineM);
        &setup_edit_dialog_usb($IMachineM);
        $gui{comboboxEditGenOSFam}->signal_handler_unblock($signal{famedit});
        $gui{comboboxEditGenOSVer}->signal_handler_unblock($signal{veredit});
        $gui{comboboxEditAudioDriver}->signal_handler_unblock($signal{audiodrv});
        $gui{comboboxEditAudioCtr}->signal_handler_unblock($signal{audioctr});
        $gui{comboboxEditNetType1}->signal_handler_unblock($signal{netadaptype1});
        $gui{comboboxEditNetType2}->signal_handler_unblock($signal{netadaptype2});
        $gui{comboboxEditNetType3}->signal_handler_unblock($signal{netadaptype3});
        $gui{comboboxEditNetType4}->signal_handler_unblock($signal{netadaptype4});
        $gui{comboboxEditNetAttach1}->signal_handler_unblock($signal{netattach1});
        $gui{comboboxEditNetAttach2}->signal_handler_unblock($signal{netattach2});
        $gui{comboboxEditNetAttach3}->signal_handler_unblock($signal{netattach3});
        $gui{comboboxEditNetAttach4}->signal_handler_unblock($signal{netattach4});
        $gui{comboboxEditNetName1}->signal_handler_unblock($signal{netname1});
        $gui{comboboxEditNetName2}->signal_handler_unblock($signal{netname2});
        $gui{comboboxEditNetName3}->signal_handler_unblock($signal{netname3});
        $gui{comboboxEditNetName4}->signal_handler_unblock($signal{netname4});
        &set_status_msg("Settings retrieved for $gname");
        my $response = $gui{dialogEdit}->run;
        $gui{dialogEdit}->hide;

        if ($response eq 'ok') {
            vboxService->IMachine_saveSettings($IMachineM);
            &get_guest_list();
        }
        else { vboxService->IMachine_discardSettings($IMachineM); }

        vboxService->ISession_close($ISession);
        $EIMachineM = undef;
        $gui{menuitemAttachFloppy} = undef; # These must be freed
        $gui{menuitemAttachDVD} = undef;
        $gui{menuitemAttachHD} = undef;
        $gui{menuAttachFloppy} = undef;
        $gui{menuAttachDVD} = undef;
        $gui{menuAttachHD} = undef;
        $gui{menuAttachAdd} = undef;
    }
    else {
        &show_error_dialog('Unable to Edit Settings',
                           'Cannot edit the settings because either the the guest is running or another program' .
                           ' already has an open session with this guest.');
    }
    &unbusy_window($gui{windowMain});
}

sub setup_edit_dialog_general() {
    my ($IMachine, $treeview) = @_;
    my $IGuestOSType = vboxService->IVirtualBox_getGuestOSType($gui{websn}, &get_selected($treeview, 4));
    $gui{entryEditGenName}->set_text(vboxService->IMachine_getName($IMachine));

    foreach my $fam (sort(keys %osfamily)) {
        my $iter = &append_liststore($gui{liststoreEditGenOSFam}, (0 => "$osfamily{$fam}{description}",
                                                                   1 => "$fam" ));
        $gui{comboboxEditGenOSFam}->set_active_iter($iter) if ($fam eq $$IGuestOSType{familyId});
    }

    foreach my $ver (sort(@{$osfamily{$$IGuestOSType{familyId}}{verids}})) {
        my $iter = &append_liststore($gui{liststoreEditGenOSVer}, (0 => "$osversion{$ver}{description}",
                                                                   1 => "$ver"));
        $gui{comboboxEditGenOSVer}->set_active_iter($iter) if ($ver eq $$IGuestOSType{id});
    }
}

sub setup_edit_dialog_system() {
    my ($IMachine) = @_;
    my $IBIOSSettings = vboxService->IMachine_getBIOSSettings($IMachine);
    if (vboxService->IMachine_getFirmwareType($IMachine) eq 'EFI') { $gui{checkbuttonEditSysEFI}->set_active(1); }
    else { $gui{checkbuttonEditSysEFI}->set_active(0); }
    $gui{spinbuttonEditSysMem}->set_range(4, $hostspec{memsize});
    $gui{spinbuttonEditSysMem}->set_value(vboxService->IMachine_getMemorySize($IMachine));
    $gui{checkbuttonEditSysVTX}->set_active($ETruth{vboxService->IMachine_getHWVirtExProperty($IMachine, 'Enabled')});
    $gui{checkbuttonEditSysNested}->set_active($ETruth{vboxService->IMachine_getHWVirtExProperty($IMachine, 'NestedPaging')});
    $gui{hscaleEditSysProcessor}->set_range($hostspec{minguestcpu}, $hostspec{maxhostcpuon} + 1); # +1 works around GTK bug
    $gui{hscaleEditSysProcessor}->set_value(vboxService->IMachine_getCPUCount($IMachine));
    $gui{checkbuttonEditSysAPIC}->set_active($ETruth{vboxService->IBIOSSettings_getIOAPICEnabled($IBIOSSettings)});
    $gui{checkbuttonEditSysPAE}->set_active($ETruth{vboxService->IMachine_getCPUProperty($IMachine, 'PAE')});
    $gui{checkbuttonEditSysUTC}->set_active($ETruth{vboxService->IMachine_getRTCUseUTC($IMachine)});
    my $pointer = vboxService->IMachine_getPointingHidType($IMachine);

    if ($pointer eq 'PS2Mouse') { $gui{comboboxEditSysPointing}->set_active(1); }
    elsif ($pointer eq 'USBMouse') { $gui{comboboxEditSysPointing}->set_active(2); }
    elsif ($pointer eq 'USBTablet') { $gui{comboboxEditSysPointing}->set_active(3); }
    elsif ($pointer eq 'ComboMouse') { $gui{comboboxEditSysPointing}->set_active(4); }
    else { $gui{comboboxEditSysPointing}->set_active(0); }

    # Default to maxbootpos+1 to mean 'not set in boot order' but this number needs to be higher than
    # true boot order numbers so the disabled devices appear at the end of the list.
    my %bootorder = (Floppy   => $hostspec{maxbootpos} + 1,
                     DVD      => $hostspec{maxbootpos} + 1,
                     HardDisk => $hostspec{maxbootpos} + 1,
                     Network  => $hostspec{maxbootpos} + 1);
    # Find boot order and set value in hash accordingly. Empty boot slots return 'Null' so skip them
    foreach (1..$hostspec{maxbootpos}) {
        my $bootdev = vboxService->IMachine_getBootOrder($IMachine,$_);
        next if ($bootdev eq 'Null');
        $bootorder{$bootdev} = $_;
    }
    # Returns hash keys sorted by value (ie boot order). Disabled devices appear at end
    foreach my $dev (sort {$bootorder{$a} cmp $bootorder{$b}} keys %bootorder) {
        if ($bootorder{$dev} == $hostspec{maxbootpos} + 1) { &append_liststore($gui{liststoreEditSysBoot}, (0 => 0, 1 => $dev)); }
        else { &append_liststore($gui{liststoreEditSysBoot}, (0 => 1, 1 => $dev)); }
    }
}

sub setup_edit_dialog_display() {
    my ($IMachine) = @_;
    my $IVRDPServer = vboxService->IMachine_getVRDPServer($IMachine);
    $gui{spinbuttonEditDispVidMem}->set_range($hostspec{minguestvram}, $hostspec{maxguestvram});
    $gui{spinbuttonEditDispVidMem}->set_value(vboxService->IMachine_getVRAMSize($IMachine));
    $gui{spinbuttonEditDispMonitor}->set_range(1, $hostspec{maxmonitors});
    $gui{spinbuttonEditDispMonitor}->set_value(vboxService->IMachine_getMonitorCount($IMachine));
    $gui{checkbuttonEditDisp3D}->set_active($ETruth{vboxService->IMachine_getAccelerate3DEnabled($IMachine)});
    $gui{checkbuttonEditDisp2D}->set_active($ETruth{vboxService->IMachine_getAccelerate2DVideoEnabled($IMachine)});
    $gui{checkbuttonEditDispServer}->set_active($ETruth{vboxService->IVRDPServer_getEnabled($IVRDPServer)});
    $gui{entryEditDispPort}->set_text(vboxService->IVRDPServer_getPorts($IVRDPServer));
    $gui{entryEditDispAuthTime}->set_text(vboxService->IVRDPServer_getAuthTimeout($IVRDPServer));
    $gui{checkbuttonEditDispMultiple}->set_active($ETruth{vboxService->IVRDPServer_getAllowMultiConnection($IVRDPServer)});
    my $authtype = vboxService->IVRDPServer_getAuthType($IVRDPServer);

    if ($authtype eq 'External') { $gui{comboboxDispAuthMeth}->set_active(1); }
    elsif ($authtype eq 'Guest') { $gui{comboboxDispAuthMeth}->set_active(2); }
    else { $gui{comboboxDispAuthMeth}->set_active(0); }
}

sub setup_edit_dialog_audio() {
    my ($IMachine) = @_;
    my $IAudioAdapter = vboxService->IMachine_getAudioAdapter($IMachine);
    $gui{checkbuttonEditAudioEnable}->set_active($ETruth{vboxService->IAudioAdapter_getEnabled($IAudioAdapter)});

    foreach my $driver (sort(keys %EAudioDriverType)) {
        if ($hostspec{os} =~ m/${EAudioDriverType{$driver}{os}}/ or $EAudioDriverType{$driver}{os} eq 'All') {
            my $iter = &append_liststore($gui{liststoreEditAudioDriver}, (0 => $EAudioDriverType{$driver}{desc}, 1 => $driver ));
            $gui{comboboxEditAudioDriver}->set_active_iter($iter) if ($driver eq vboxService->IAudioAdapter_getAudioDriver($IAudioAdapter));
        }
    }

    foreach my $ctr (sort(keys %EAudioControllerType)) {
        my $iter = &append_liststore($gui{liststoreEditAudioCtr}, (0 => $EAudioControllerType{$ctr}, 1 => $ctr ));
        $gui{comboboxEditAudioCtr}->set_active_iter($iter) if ($ctr eq vboxService->IAudioAdapter_getAudioController($IAudioAdapter));
    }
}

sub setup_edit_dialog_network() {
    my ($IMachine) = @_;
    my $IHost = vboxService->IVirtualBox_getHost($gui{websn});

    foreach my $slot (1..4) {
        my $INetworkAdapter = vboxService->IMachine_getNetworkAdapter($IMachine,($slot-1));
        my $NetAdapterType = vboxService->INetworkAdapter_getAdapterType($INetworkAdapter);
        my $AttachType = vboxService->INetworkAdapter_getAttachmentType($INetworkAdapter);
        my $hostinterface = vboxService->INetworkAdapter_getHostInterface($INetworkAdapter);
        my @IHostNetworkInterface = vboxService->IHost_getNetworkInterfaces($IHost);
        $gui{'checkbuttonEditNetEnable' . $slot}->set_active($ETruth{vboxService->INetworkAdapter_getEnabled($INetworkAdapter)});
        $gui{'checkbuttonEditNetCable' . $slot}->set_active($ETruth{vboxService->INetworkAdapter_getCableConnected($INetworkAdapter)});
        $gui{'entryEditNetMac' . $slot}->set_text(vboxService->INetworkAdapter_getMACAddress($INetworkAdapter));
        # Set current adapter type for NICs 1-4
        foreach my $adaptertype (sort(keys %ENetworkAdapterType)) {
            my $iter = &append_liststore($gui{'liststoreEditNetAdapterType' . $slot}, (0 => $ENetworkAdapterType{$adaptertype},
                                                                                       1 => $adaptertype ));
            $gui{'comboboxEditNetType' . $slot}->set_active_iter($iter) if ($adaptertype eq $NetAdapterType);
        }
        # Set current attachment type for NICs 1-4
        foreach my $attach (sort(keys %ENetworkAttachmentType)) {
            my $iter = &append_liststore($gui{'liststoreEditNetAttachTo' . $slot}, (0 => $ENetworkAttachmentType{$attach},
                                                                                    1 => $attach));
            $gui{'comboboxEditNetAttach' . $slot}->set_active_iter($iter) if ($attach eq $AttachType);
        }

        if ($AttachType eq 'Bridged') {
            # Set current bridged network name for NICs 1-4, depending on attachment type
            foreach my $interface (@IHostNetworkInterface) {
                if (vboxService->IHostNetworkInterface_getInterfaceType($interface) eq 'Bridged') {
                    my $name = vboxService->IHostNetworkInterface_getName($interface);
                    my $iter = &append_liststore($gui{'liststoreEditNetName' . $slot}, (0 => $name));
                    $gui{'comboboxEditNetName' . $slot}->set_active_iter($iter) if ($name eq $hostinterface);
                }
            }
            &hide_off($gui{'labelEditNetName' . $slot}, $gui{'comboboxEditNetName' . $slot});
            &hide_on($gui{'labelEditNetNameInt' . $slot}, $gui{'entryEditNetNameInt' . $slot});
        }
        elsif ($AttachType eq 'HostOnly') {
            # Set current hostonly network name for NICs 1-4, depending on attachment type
            foreach my $interface (@IHostNetworkInterface) {
                if (vboxService->IHostNetworkInterface_getInterfaceType($interface) eq 'HostOnly') {
                    my $name = vboxService->IHostNetworkInterface_getName($interface);
                    my $iter = &append_liststore($gui{'liststoreEditNetName' . $slot}, (0 => $name));
                    $gui{'comboboxEditNetName' . $slot}->set_active_iter($iter) if ($name eq $hostinterface);
                }
            }
            &hide_off($gui{'labelEditNetName' . $slot}, $gui{'comboboxEditNetName' . $slot});
            &hide_on($gui{'labelEditNetNameInt' . $slot}, $gui{'entryEditNetNameInt' . $slot});
        }
        elsif ($AttachType eq 'Internal') {
            my $name = vboxService->INetworkAdapter_getInternalNetwork($INetworkAdapter);
            $gui{'entryEditNetNameInt' . $slot}->set_text($name);
            &hide_on($gui{'labelEditNetName' . $slot}, $gui{'comboboxEditNetName' . $slot});
            &hide_off($gui{'labelEditNetNameInt' . $slot}, $gui{'entryEditNetNameInt' . $slot});
        }
        else {
            # All other networking types do not have name associated with them
            &hide_on($gui{'labelEditNetName' . $slot}, $gui{'comboboxEditNetName' . $slot},
                     $gui{'labelEditNetNameInt' . $slot}, $gui{'entryEditNetNameInt' . $slot});
        }
    }
}

sub setup_edit_dialog_usb() {
    my ($IMachine) = @_;
    my $IUSBController = vboxService->IMachine_getUSBController($EIMachineM);
    $gui{checkbuttonEditUSBEnable}->set_active($ETruth{vboxService->IUSBController_getEnabled($IUSBController)});
    $gui{checkbuttonEditUSBEHCI}->set_active($ETruth{vboxService->IUSBController_getEnabledEhci($IUSBController)});
}

sub setup_edit_dialog_storage() {
    my ($IMachine) = @_;
    &tgr_editstor_tree_fill($IMachine);
    $gui{menuAttachAdd} = Gtk2::Menu->new();
    $gui{menuAttachHD} = Gtk2::Menu->new();
    $gui{menuAttachDVD} = Gtk2::Menu->new();
    $gui{menuAttachFloppy} = Gtk2::Menu->new();
    $gui{menuitemAttachHD} = Gtk2::MenuItem->new('Hard Disks');
    $gui{menuitemAttachDVD} = Gtk2::MenuItem->new('CDs/DVDs');
    $gui{menuitemAttachFloppy} = Gtk2::MenuItem->new('Floppy Disks');
    $gui{menuitemAttachHD}->set_submenu($gui{menuAttachHD});
    $gui{menuitemAttachDVD}->set_submenu($gui{menuAttachDVD});
    $gui{menuitemAttachFloppy}->set_submenu($gui{menuAttachFloppy});
    $gui{menuitemAttachHD}->show();
    $gui{menuitemAttachDVD}->show();
    $gui{menuitemAttachFloppy}->show();
    $gui{menuAttachAdd}->append($gui{menuitemAttachHD});
    $gui{menuAttachAdd}->append($gui{menuitemAttachDVD});
    $gui{menuAttachAdd}->append($gui{menuitemAttachFloppy});

    my @IMediumHD = vboxService->IVirtualBox_getHardDisks($gui{websn});
    my @IMediumDVD = vboxService->IVirtualBox_getDVDImages($gui{websn});
    my @IMediumFloppy = vboxService->IVirtualBox_getFloppyImages($gui{websn});

    foreach my $IMedium (@IMediumHD) {
        my $name = vboxService->IMedium_getName($IMedium);
        my $item = Gtk2::MenuItem->new($name);
        $gui{menuAttachHD}->append($item);
        $item->show();
        $item->signal_connect(activate => \&tgr_editstor_attach_add, $IMedium);
    }

    foreach my $IMedium (@IMediumDVD) {
        my $name = vboxService->IMedium_getName($IMedium);
        my $item = Gtk2::MenuItem->new($name);
        $gui{menuAttachDVD}->append($item);
        $item->show();
        $item->signal_connect(activate => \&tgr_editstor_attach_add, $IMedium);
    }

    foreach my $IMedium (@IMediumFloppy) {
        my $name = vboxService->IMedium_getName($IMedium);
        my $item = Gtk2::MenuItem->new($name);
        $gui{menuAttachFloppy}->append($item);
        $item->show();
        $item->signal_connect(activate => \&tgr_editstor_attach_add, $IMedium);
    }
}

sub get_guest_list() {
    &busy_window($gui{windowMain});
    &sensitive_off(@{$sensgrp{unselected}});
    &set_status_msg("Retrieving the list of guests from $endpoint...");
    &clear_liststore($gui{liststoreGuest}, $gui{liststoreDetails});
    $gui{textbufferDescription}->set_text('');
    my @IMachine = vboxService->IVirtualBox_getMachines($gui{websn});

    foreach my $machine (@IMachine) {
        my $gname = vboxService->IMachine_getName($machine);
        my $status = vboxService->IMachine_getAccessible($machine);
        &set_status_msg("Processing guest: $gname...");

        if ($status eq 'false') {
            my $IVirtualBoxErrorInfo = vboxService->IMachine_getAccessError($machine);
            $status = vboxService->IVirtualBoxErrorInfo_getText($IVirtualBoxErrorInfo);
        }
        else { $status = vboxService->IMachine_getState($machine); }

        &append_liststore($gui{liststoreGuest}, (0 => $gname,
                                                 1 => $osversion{vboxService->IMachine_getOSTypeId($machine)}{description},
                                                 2 => $machine,
                                                 3 => $status,
                                                 4 => vboxService->IMachine_getOSTypeId($machine),
                                                 5 => vboxService->IMachine_getId($machine)));
    }

    &set_status_msg("Retrieved list of guests from $endpoint.");
    &unbusy_window($gui{windowMain});
}

sub open_remote_display() {
    my ($widget, $treeview) = @_;
    my ($name, $os, $IMachine) = &get_selected($treeview);
    my $IVRDPServer = vboxService->IMachine_getVRDPServer($IMachine);
    my $rdpcmd = $prefs{RDPCLIENT};
    my $pass = $gui{entryConnectPassword}->get_text();
    my $user = $gui{comboboxentryConnectUser}->get_active_text();

    if (vboxService->IVRDPServer_getEnabled($IVRDPServer) eq 'true') {
        my @ports = split ( ',', vboxService->IVRDPServer_getPorts($IVRDPServer)); # Multiple ports may be return ',' separated
        my $dst = $endpoint;
        $dst =~ s/^.*:\/\///;
        $dst =~ s/:\d+$//;
        $rdpcmd =~ s/%h/$dst/g;
        $rdpcmd =~ s/%p/$ports[0]/g;
        $rdpcmd =~ s/%n/$name/g;
        $rdpcmd =~ s/%o/$os/g;
        $rdpcmd =~ s/%U/$user/g;
        $rdpcmd =~ s/%P/$pass/g;
        system("$rdpcmd &");
        &set_status_msg("Sent request to open remote display: $dst:$ports[0]");
    }
    else {
        &show_error_dialog('Remote Display Not Enabled',
                           'The remote display server for this guest' .
                           ' has not been enabled or is not running.' .
                           ' Please enable it in the guest settings.');
    }
}

sub populate_hostspec() {
    my $IHost = vboxService->IVirtualBox_getHost($gui{websn});
    my $ISystemProperties = vboxService->IVirtualBox_getSystemProperties($gui{websn});
    %hostspec = (vbver        => vboxService->IVirtualBox_getVersion($gui{websn}),
                 buildrev     => vboxService->IVirtualBox_getRevision($gui{websn}),
                 pkgtype      => vboxService->IVirtualBox_getPackageType($gui{websn}),
                 os           => vboxService->IHost_getOperatingSystem($IHost),
                 osver        => vboxService->IHost_getOSVersion($IHost),
                 settingsfile => vboxService->IVirtualBox_getSettingsFilePath($gui{websn}),
                 machinedir   => vboxService->ISystemProperties_getDefaultMachineFolder($ISystemProperties),
                 hddir        => vboxService->ISystemProperties_getDefaultHardDiskFolder($ISystemProperties),
                 memsize      => vboxService->IHost_getMemorySize($IHost),
                 maxhdsize    => vboxService->ISystemProperties_getMaxVDISize($ISystemProperties),
                 maxnet       => vboxService->ISystemProperties_getNetworkAdapterCount($ISystemProperties),
                 maxser       => vboxService->ISystemProperties_getSerialPortCount($ISystemProperties),
                 hdfolder     => vboxService->ISystemProperties_getDefaultHardDiskFolder($ISystemProperties),
                 minguestcpu  => vboxService->ISystemProperties_getMinGuestCPUCount($ISystemProperties),
                 maxguestcpu  => vboxService->ISystemProperties_getMaxGuestCPUCount($ISystemProperties),
                 maxhostcpuon => vboxService->IHost_getProcessorOnlineCount($IHost),
                 cpudesc      => vboxService->IHost_getProcessorDescription($IHost),
                 cpuspeed     => vboxService->IHost_getProcessorSpeed($IHost),
                 minguestram  => vboxService->ISystemProperties_getMinGuestRAM($ISystemProperties),
                 maxguestram  => vboxService->ISystemProperties_getMaxGuestRAM($ISystemProperties),
                 minguestvram => vboxService->ISystemProperties_getMinGuestVRAM($ISystemProperties),
                 maxguestvram => vboxService->ISystemProperties_getMaxGuestVRAM($ISystemProperties),
                 maxbootpos   => vboxService->ISystemProperties_getMaxBootPosition($ISystemProperties),
                 maxmonitors  => vboxService->ISystemProperties_getMaxGuestMonitors($ISystemProperties));
}

sub populate_ostypes() {
    my @IGuestOSType = vboxService->IVirtualBox_getGuestOSTypes($gui{websn});
    %osfamily=();
    %osversion=();

    foreach my $type (@IGuestOSType) {
        if (!defined($osfamily{$$type{familyId}})) {
            $osfamily{$$type{familyId}} = {};
            $osfamily{$$type{familyId}}{verids} = ();
        }

        $osfamily{$$type{familyId}}{description} = $$type{familyDescription};
        push @{ $osfamily{$$type{familyId}}{verids} }, $$type{id};
        $osversion{$$type{id}} = {} if (!defined($osversion{$$type{id}}));
        $osversion{$$type{id}}{description} = $$type{description};
        $osversion{$$type{id}}{adapterType} = $$type{adapterType};
        $osversion{$$type{id}}{recommendedHDD} = $$type{recommendedHDD};
        $osversion{$$type{id}}{is64Bit} = $$type{is64Bit};
        $osversion{$$type{id}}{recommendedVirtEx} = $$type{recommendedVirtEx};
        $osversion{$$type{id}}{recommendedIOAPIC} = $$type{recommendedIOAPIC};
        $osversion{$$type{id}}{recommendedVRAM} = $$type{recommendedVRAM};
        $osversion{$$type{id}}{recommendedRAM} = $$type{recommendedRAM};
        $osversion{$$type{id}}{recommendedHpet} = $$type{recommendedHpet};
        $osversion{$$type{id}}{recommendedUsbHid} = $$type{recommendedUsbHid};
        $osversion{$$type{id}}{recommendedVirtEx} = $$type{recommendedVirtEx};
        $osversion{$$type{id}}{recommendedPae} = $$type{recommendedPae};
        $osversion{$$type{id}}{recommendedUsbTablet} = $$type{recommendedUsbTablet};
        $osversion{$$type{id}}{recommendedHdStorageBus} = $$type{recommendedHdStorageBus};
        $osversion{$$type{id}}{recommendedFirmware} = $$type{recommendedFirmware};
        $osversion{$$type{id}}{recommendedDvdStorageBus} = $$type{recommendedDvdStorageBus};
        $osversion{$$type{id}}{recommendedHdStorageController} = $$type{recommendedHdStorageController};
        $osversion{$$type{id}}{recommendedDvdStorageController} = $$type{recommendedDvdStorageController};
        $osversion{$$type{id}}{recommendedRtcUseUtc} = $$type{recommendedRtcUseUtc};
        $osversion{$$type{id}}{familyId} = $$type{familyId};
    }
}

# Need to report an error if mutable session not obtained
sub set_guest_description() {
    my ($widget, $treeview) = @_;
    my $IMachine = &get_selected($treeview, 2);
    my ($ISession, $IMachineM) = &get_mutable_session($IMachine);
    if ($IMachineM) {
        &set_status_msg('Saving description...');
        my $iter_s = $gui{textbufferDescription}->get_start_iter();
        my $iter_e = $gui{textbufferDescription}->get_end_iter();
        vboxService->IMachine_setDescription($IMachineM, $gui{textbufferDescription}->get_text($iter_s, $iter_e, 0));
        vboxService->IMachine_saveSettings($IMachineM);
        vboxService->ISession_close($ISession);
        &set_status_msg('Description saved.');
    }
    else {
        &show_error_dialog('Guest State is Immutable',
                           'The description cannot be saved because either the guest is running or another' .
                           ' program already has an open session with this guest.');
    }
}

sub create_new_guest() {
    my ($name, $fam, $ver, $mem) = @_;
    my $IMachine;
    my $dvdctrname = $osversion{$ver}{recommendedDvdStorageBus} .  ' Controller';
    my $hdctrname = $osversion{$ver}{recommendedHdStorageBus} . ' Controller';
    &set_status_msg("Creating new guest $name...");
    my $portnum = int(rand(15000)) + 50000; # Pick random between 50000 and 65000
    $IMachine = vboxService->IVirtualBox_createMachine($gui{websn}, $name, $ver, '', 'UUID 00000000-0000-0000-0000-000000000000');

    if ($IMachine) {
        my $IVRDPServer = vboxService->IMachine_getVRDPServer($IMachine);
        my $IBIOSSettings = vboxService->IMachine_getBIOSSettings($IMachine);
        vboxService->IBIOSSettings_setIOAPICEnabled($IBIOSSettings, $ETruth{$osversion{$ver}{recommendedIOAPIC}});
        vboxService->IMachine_setCPUProperty($IMachine, 'PAE', $ETruth{$osversion{$ver}{recommendedPae}});
        vboxService->IMachine_setHWVirtExProperty($IMachine, 'Enabled', $ETruth{$osversion{$ver}{recommendedVirtEx}});
        vboxService->IMachine_setVRAMSize($IMachine, $osversion{$ver}{recommendedVRAM});
        vboxService->IMachine_setFirmwareType($IMachine, $osversion{$ver}{recommendedFirmware});
        vboxService->IMachine_setMemorySize($IMachine, $mem);
        my $IStorCtrDVD = vboxService->IMachine_addStorageController($IMachine, $dvdctrname, $osversion{$ver}{recommendedDvdStorageBus});
        vboxService->IStorageController_setControllerType($IStorCtrDVD, $osversion{$ver}{recommendedDvdStorageController});

        if ($osversion{$ver}{recommendedHdStorageBus} ne $osversion{$ver}{recommendedDvdStorageBus}) {
            my $IStorCtrHD = vboxService->IMachine_addStorageController($IMachine, $hdctrname, $osversion{$ver}{recommendedHdStorageBus});
            vboxService->IStorageController_setControllerType($IStorCtrHD, $osversion{$ver}{recommendedHdStorageController});
        }

        vboxService->IVRDPServer_setPorts($IVRDPServer, $portnum);
        vboxService->IVRDPServer_setEnabled($IVRDPServer, 'true');
        vboxService->IMachine_saveSettings($IMachine);
        vboxService->IVirtualBox_registerMachine($gui{websn}, $IMachine);
        &set_status_msg("Completed creation of guest $name.");
    }

    return $IMachine, $dvdctrname, $hdctrname;
}

sub create_new_HD() {
    my ($diskname, $format, $type, $size) = @_;
    &set_status_msg("Creating new hard disk...");
    my $IMedium = vboxService->IVirtualBox_createHardDisk($gui{websn}, $format, "$hostspec{hdfolder}/$diskname.vdi");
    my $IProgress = vboxService->IMedium_createBaseStorage($IMedium, $size, $type);

    do {
        &set_status_msg('Creating new hard disk... ' . vboxService->IProgress_getPercent($IProgress) . '% complete.');
        sleep 1; # Poll only every 1 second
    } until (vboxService->IProgress_getCompleted($IProgress) eq 'true');

    &set_status_msg('Completed creation of hard disk.');
    return $IMedium;
}

# SATA/SAS Controllers need a port created before attachments can be made
sub create_sas_sata_port() {
    my ($IStorCtr) = @_;
    my $bus = vboxService->IStorageController_getBus($IStorCtr);
    if ($bus eq 'SATA' or $bus eq 'SAS') {
        my $pc = vboxService->IStorageController_getPortCount($IStorCtr);
        my $mpc = vboxService->IStorageController_getMaxPortCount($IStorCtr);
        vboxService->IStorageController_setPortCount($IStorCtr, $pc + 1) if ($pc < $mpc);
    }
}

sub tgr_keyboard_CAD() {
    my ($widget, $treeview) = @_;
    my $gname = &get_selected($treeview, 0);
    my $guuid = &get_selected($treeview, 5);
    my $IConsole = &get_IConsole($guuid);
    my $IKeyboard = vboxService->IConsole_getKeyboard($IConsole);
    vboxService->IKeyboard_putCAD($IKeyboard);
    &set_status_msg("Sent Ctr-Alt-Del to $gname");
}


sub tgr_newgen_osfam() {
    my ($combofam, $combover) = @_;
    my $fam = &get_selected($combofam, 1);
    $combofam->signal_handler_block($signal{fam}); # Block to avoid signal emission when changing
    $combover->signal_handler_block($signal{ver});
    &clear_liststore($gui{liststoreNewOSVer});

    foreach my $ver (sort(@{$osfamily{$fam}{verids}})) {
        my $iter = &append_liststore($gui{liststoreNewOSVer}, (0 => "$osversion{$ver}{description}",
                                                               1 => "$ver" ));
        $combover->set_active_iter($iter) if ($ver eq 'WindowsXP' | $ver eq 'Ubuntu' | $ver eq 'Solaris');
    }

    $combover->set_active(0) if ($combover->get_active() == -1);
    $combofam->signal_handler_unblock($signal{fam});
    $combover->signal_handler_unblock($signal{ver});
    $combover->signal_emit('changed'); # Force update of other fields based on OS
}

sub tgr_newgen_osver() {
    my ($combover, $combofam) = @_;
    my $ver = &get_selected($combover, 1);
    $combofam->signal_handler_block($signal{fam}); # Avoid signal emission when changing
    $combover->signal_handler_block($signal{ver});
    $gui{spinbuttonNewMemory}->set_value($osversion{$ver}{recommendedRAM});
    $gui{spinbuttonNewHDSize}->set_value($osversion{$ver}{recommendedHDD});
    $combofam->signal_handler_unblock($signal{fam});
    $combover->signal_handler_unblock($signal{ver});
}

sub tgr_newstor_new_exist() {
    my ($widget) = @_;
    my $buttongrp = $widget->get_group();

    if ($$buttongrp[0]->get_active() == 1) {
        &sensitive_on($gui{comboboxNewChooseHD});
        &sensitive_off($gui{checkbuttonNewDynamic}, $gui{spinbuttonNewHDSize});
    }
    else {
        &sensitive_off($gui{comboboxNewChooseHD});
        &sensitive_on($gui{checkbuttonNewDynamic}, $gui{spinbuttonNewHDSize});
    }
}

sub show_tg_menu() {
    my ($widget, $event) = @_;
    $gui{menuTG}->popup(undef, undef, undef, undef, 0, $event->time) if ($event->button == 3);
    return 0;
}

sub tgr_guestlist_selected() {
    my ($widget, $liststore) = @_;
    my $IMachine = &get_selected($gui{treeviewGuest}, 2);
    &clear_liststore($liststore, $gui{treestoreSnapshots});
    &sensitive_off(@{$sensgrp{unselected}});
    $gui{textbufferDescription}->set_text(vboxService->IMachine_getDescription($IMachine));
    my $status = vboxService->IMachine_getState($IMachine);

    if ($status eq 'Running' | $status eq 'Starting') { &sensitive_on(@{$sensgrp{running}}); }
    elsif ($status eq 'Saved' | $status eq 'Paused') { &sensitive_on(@{$sensgrp{saved}}); }
    elsif ($status eq 'PoweredOff' | $status eq 'Aborted') { &sensitive_on(@{$sensgrp{poweroff}}); }
    else { &sensitive_off(@{$sensgrp{unselected}}); }
}

sub tgr_guest_double() {
    my ($widget) = @_;
    &get_snapshots($widget, $gui{treestoreSnapshots});
    &get_details($widget, $gui{liststoreDetails});
}

sub tgr_editgen_name() {
    my ($widget) = @_;
    my $name = $widget->get_text();
    vboxService->IMachine_setName($EIMachineM, $name) if ($name);
}

sub tgr_editgen_osfam() {
    my ($combofam, $combover) = @_;
    my $fam = &get_selected($combofam, 1);
    $combofam->signal_handler_block($signal{famedit}); # Block to avoid signal emission when changing
    $combover->signal_handler_block($signal{veredit});
    &clear_liststore($gui{liststoreEditGenOSVer});

    foreach my $ver (sort(@{$osfamily{$fam}{verids}})) {
        my $iter = &append_liststore($gui{liststoreEditGenOSVer}, (0 => "$osversion{$ver}{description}",
                                                                   1 => "$ver" ));
        $combover->set_active_iter($iter) if ($ver eq 'WindowsXP' | $ver eq 'Ubuntu' | $ver eq 'Solaris');
    }

    $combover->set_active(0) if ($combover->get_active() == -1);
    $combofam->signal_handler_unblock($signal{famedit});
    $combover->signal_handler_unblock($signal{veredit});
    $combover->signal_emit('changed'); # Force update of other fields based on OS
}

sub tgr_editgen_osver() {
    my ($combover) = @_;
    my $ver = &get_selected($combover, 1);
    vboxService->IMachine_setOSTypeId($EIMachineM, $ver);
}

sub tgr_editsys_boot_toggle() {
    my ($widget, $path_str, $model) = @_;
    my $iter = $model->get_iter(Gtk2::TreePath->new_from_string($path_str));
    my $val = $model->get($iter, 0);
    $model->set ($iter, 0, !$val); # Always set the opposite of val to act as a toggle
}

sub tgr_editsys_mem() {
    my ($widget, $focus, $adjust) = @_;
    vboxService->IMachine_setMemorySize($EIMachineM, int($adjust->get_value()));
    return 0; # Must return this value for the signal used.
}

sub tgr_editsys_apic() {
    my ($widget, $focus) = @_;
    my $IBIOSSettings = vboxService->IMachine_getBIOSSettings($EIMachineM);
    vboxService->IBIOSSettings_setIOAPICEnabled($IBIOSSettings, $widget->get_active());
}

sub tgr_editsys_efi() {
    my ($widget, $focus) = @_;
    if ($widget->get_active() == 1) { vboxService->IMachine_setFirmwareType($EIMachineM, 'EFI'); }
    else { vboxService->IMachine_setFirmwareType($EIMachineM, 'BIOS'); }
}

sub tgr_editsys_processors() {
    my ($widget, $focus, $adjust) = @_;
    vboxService->IMachine_setCPUCount($EIMachineM, int($adjust->get_value()));
    return 0; # Must return this value for the signal used.
}

sub tgr_editsys_pae() {
    my ($widget, $focus) = @_;
    vboxService->IMachine_setCPUProperty($EIMachineM, 'PAE', $widget->get_active());
}

sub tgr_editsys_vtx() {
    my ($widget, $focus) = @_;
    vboxService->IMachine_setHWVirtExProperty($EIMachineM, 'Enabled', $widget->get_active());
}

sub tgr_editsys_utc() {
    my ($widget, $focus) = @_;
    vboxService->IMachine_setRTCUseUTC($EIMachineM, $widget->get_active());
}

sub tgr_editsys_pointer() {
    my ($widget) = @_;
    vboxService->IMachine_setPointingHidType($EIMachineM, &get_selected($widget, 1));
}

sub tgr_editsys_nested() {
    my ($widget, $focus) = @_;
    vboxService->IMachine_setHWVirtExProperty($EIMachineM, 'NestedPaging', $widget->get_active());
}

sub tgr_editsys_boot_up() {
    my ($widget, $treeview) = @_;
    my ($model, $iter) = $treeview->get_selection->get_selected;
    my $path = $model->get_path($iter);
    $path->prev;
    my $previter = $model->get_iter($path);
    $model->move_before($iter, $previter);
}

sub tgr_editsys_boot_down() {
    my ($widget, $treeview) = @_;
    my ($model, $iter) = $treeview->get_selection->get_selected;
    my $nextiter = $model->iter_next($iter);
    $model->move_after($iter, $nextiter) if ($nextiter);
}

sub tgr_editsys_bootorder() {
    my ($widget, $treeview) = @_;
    my $model = $treeview->get_model();
    my $iter = $model->get_iter_first();
    my $i = 1;

    while ($iter) {
        my $dev = $model->get($iter,1);
        $dev = 'Null' if ($model->get($iter,0) == 0);
        vboxService->IMachine_setBootOrder($EIMachineM, $i, $dev);
        $iter = $model->iter_next($iter);
        $i++;
    }
}

sub tgr_editsys_boot_selected() {
    &sensitive_on($gui{buttonEditSysBootDown}, $gui{buttonEditSysBootUp});
}

sub tgr_editdisp_vidmem() {
    my ($widget, $focus, $adjust) = @_;
    vboxService->IMachine_setVRAMSize($EIMachineM, int($adjust->get_value()));
    return 0; # Must return this value for the signal used.
}

sub tgr_editdisp_monitors() {
    my ($widget, $focus, $adjust) = @_;
    vboxService->IMachine_setMonitorCount($EIMachineM, int($adjust->get_value()));
    return 0; # Must return this value for the signal used.
}

sub tgr_editdisp_2D() {
    my ($widget, $focus) = @_;
    vboxService->IMachine_setAccelerate2DVideoEnabled($EIMachineM, $widget->get_active());
}

sub tgr_editdisp_3D() {
    my ($widget, $focus) = @_;
    vboxService->IMachine_setAccelerate3DEnabled($EIMachineM, $widget->get_active());
}

sub tgr_editdisp_RDP() {
    my ($widget, $focus) = @_;
    my $IVRDPServer = vboxService->IMachine_getVRDPServer($EIMachineM);
    vboxService->IVRDPServer_setEnabled($IVRDPServer, $widget->get_active());
}

sub tgr_editdisp_RDPTime() {
    my ($widget) = @_;
    my $authtime = int($widget->get_text());
    my $IVRDPServer = vboxService->IMachine_getVRDPServer($EIMachineM);
    vboxService->IVRDPServer_setAuthTimeout($IVRDPServer, $authtime) if ($authtime);
}

sub tgr_editdisp_RDPPort() {
    my ($widget) = @_;
    my $ports = $widget->get_text();
    # Ideally some additional validation would be needed here
    my $IVRDPServer = vboxService->IMachine_getVRDPServer($EIMachineM);
    vboxService->IVRDPServer_setPorts($IVRDPServer, $ports) if ($ports);
}

sub tgr_editdisp_RDPAuth() {
    my ($widget) = @_;
    my $IVRDPServer = vboxService->IMachine_getVRDPServer($EIMachineM);
    vboxService->IVRDPServer_setAuthType($IVRDPServer, &get_selected($widget, 1));
}

sub tgr_editdisp_RDPMulti() {
    my ($widget, $focus) = @_;
    my $IVRDPServer = vboxService->IMachine_getVRDPServer($EIMachineM);
    vboxService->IVRDPServer_setAllowMultiConnection($IVRDPServer, $widget->get_active());
}

sub tgr_editstor_tree_fill() {
    my ($IMachine) = @_;
    &busy_window($gui{dialogEdit});
    &sensitive_off(@{$sensgrp{stor_unsel}});
    &clear_liststore($gui{treestoreEditStor});
    my @IStorageController = vboxService->IMachine_getStorageControllers($IMachine);

    foreach my $controller (@IStorageController) {
        my $name = vboxService->IStorageController_getName($controller);
        my $bus = vboxService->IStorageController_getBus($controller);
        my $iter = $gui{treestoreEditStor}->append(undef);
        $gui{treestoreEditStor}->set($iter, (0 => $name,
                                             1 => $name,
                                             2 => $bus . ' Controller',
                                             3 => $controller));
        my @IMediumAttachment = vboxService->IMachine_getMediumAttachmentsOfController($IMachine, $name);

        foreach my $attachment (@IMediumAttachment) {
            my $citer = $gui{treestoreEditStor}->append($iter);
            ## Needs fixing - does it???
            if ($$attachment{medium}) {
                $gui{treestoreEditStor}->set($citer, (0 => vboxService->IMedium_getName($$attachment{medium}),
                                                      1 => $name,
                                                      2 => $$attachment{type},
                                                      3 => $$attachment{medium},
                                                      4 => $$attachment{device},
                                                      5 => $$attachment{port},
                                                      6 => vboxService->IMedium_getLocation($$attachment{medium})));
            }
            else {
                $gui{treestoreEditStor}->set($citer, (0 => '<empty>',
                                                      1 => $name,
                                                      2 => $$attachment{type},
                                                      3 => '<empty>',
                                                      4 => $$attachment{device},
                                                      5 => $$attachment{port},
                                                      6 => '<empty>'));
            }
        }
    }
    $gui{treeviewEditStor}->expand_all();
    &unbusy_window($gui{dialogEdit});
}

sub show_ctr_menu() {
    my ($widget, $event) = @_; #$event->time
    $gui{menuCtrAdd}->popup(undef, undef, undef, undef, 0, $event->time) if ($event->button == 1);
    return 0;
}

sub tgr_editstor_ctr_add() {
    my ($widget) = @_;
    my $bus = 'Floppy'; # Assume floppy unless set otherwise

    if ($widget eq $gui{menuitemCtrAddIDE}) { $bus = 'IDE'; }
    elsif ($widget eq $gui{menuitemCtrAddSCSI}) { $bus = 'SCSI'; }
    elsif ($widget eq $gui{menuitemCtrAddSATA}) { $bus = 'SATA'; }
    elsif ($widget eq $gui{menuitemCtrAddSAS}) { $bus = 'SAS'; }

    my @IStorageController = vboxService->IMachine_getStorageControllers($EIMachineM);
    my $exists = 0;

    foreach my $ctr (@IStorageController) {
        $exists = 1 if (vboxService->IStorageController_getBus($ctr) eq $bus)
    }

    if (!$exists) {
        vboxService->IMachine_addStorageController($EIMachineM, "$bus Controller", $bus);
        &tgr_editstor_tree_fill($EIMachineM);
    }
    elsif ($exists) {
        &show_error_dialog('Controller Already Allocated',
                           'The guest already has a storage controller of this type assigned to it.');
    }
}

sub show_attach_menu() {
    my ($widget, $event) = @_;
    $gui{menuAttachAdd}->popup(undef, undef, undef, undef, 0, $event->time) if ($event->button == 1);
    return 0;
}

sub tgr_editstor_attach_add() {
    my ($widget, $IMedium) = @_;
    my $mtype = vboxService->IMedium_getDeviceType($IMedium);
    my ($cname, $cname2, $ctype, $IStorCtr) = &get_selected($gui{treeviewEditStor});
    my ($freedev, $freeport) = &get_free_deviceport($EIMachineM, $IStorCtr);
    my @IMediumAttachment = vboxService->IMachine_getMediumAttachmentsOfController($EIMachineM, $cname);
    my $muuid = vboxService->IMedium_getId($IMedium);
    my $attached = 0;

    # Check for an empty Floppy/DVD first and try an use that (if medium appropriate)
    if ($mtype eq 'DVD') {
        foreach my $attach (@IMediumAttachment) {
            next if ($$attach{medium}); # We ignore if it already has a medium
            next if ($$attach{type} ne 'DVD'); # Only interested in DVD devices
            vboxService->IMachine_mountMedium($EIMachineM, $cname, $$attach{port}, $$attach{device}, $muuid);
            $attached = 1;
        }
    }
    elsif ($mtype eq 'Floppy') {
        foreach my $attach (@IMediumAttachment) {
            next if ($$attach{medium}); # We ignore if it already has a medium
            next if ($$attach{type} ne 'Floppy'); # Only interested in DVD devices
            vboxService->IMachine_mountMedium($EIMachineM, $cname, $$attach{port}, $$attach{device}, $muuid);
            $attached = 1;
        }

    }

    # For non removable media or if no empty Floppy/DVD, then add new attachment
    if ($attached == 0) {
        if ($freedev ne -1) {
            &create_sas_sata_port($IStorCtr); # Harmless if not SAS/SATA controller
            vboxService->IMachine_attachDevice($EIMachineM, $cname, $freeport, $freedev, $mtype, $muuid);
            $attached = 1
        }
    }

    # If medium wasn't attached, controller must be full
    if (!$attached) { &show_error_dialog('Controller Full', 'The selected controller is full. No more media can be attached.'); }
    else { &tgr_editstor_tree_fill($EIMachineM); }
}

sub tgr_editstor_attach_rem() {
    my ($widget, $treeview) = @_;
    my ($mname, $cname, $mtype, $IMedium, $mdev, $mport) = &get_selected($treeview);

    # Unmount removable media, otherwise remove device for everything.
    if ((($mtype eq 'Floppy') or ($mtype eq 'DVD')) and ($IMedium ne '<empty>')) {
        vboxService->IMachine_mountMedium($EIMachineM, $cname, $mport, $mdev, 'UUID 00000000-0000-0000-0000-000000000000');
    }
    else { vboxService->IMachine_detachDevice($EIMachineM, $cname, $mport, $mdev); }

    &tgr_editstor_tree_fill($EIMachineM);
}

sub tgr_editstor_ctr_rem() {
    my ($widget, $treeview) = @_;
    my $ctrname = &get_selected($treeview, 0);
    my @IMediumAttachments = vboxService->IMachine_getMediumAttachmentsOfController($EIMachineM, $ctrname);

    if (@IMediumAttachments) {
        &show_warn_dialog('Controller in Use',
                          'The controller still has media attached. Remove media before removing controller');
    }
    else { vboxService->IMachine_removeStorageController($EIMachineM, $ctrname); }
    &tgr_editstor_tree_fill($EIMachineM);
}

sub tgr_editstor_selected() {
    my ($widget, $entryctr) = @_;
    my ($cname, $cname2, $ctype, $IStorCtr) = &get_selected($widget);
    if ($ctype =~ m/Controller/) {
        &sensitive_on(@{$sensgrp{ctr_sel_on}});
        &sensitive_off($gui{buttonEditStorRemoveAttach});
        $entryctr->set_text($cname);
        &tgr_editstor_ctrtype_fill($ctype, $IStorCtr);

        if ($ctype eq 'IDE Controller') {
            &sensitive_on($gui{menuitemAttachHD}, $gui{menuitemAttachDVD});
            &sensitive_off($gui{menuitemAttachFloppy});
        }
        elsif ($ctype eq 'Floppy Controller') {
            &sensitive_off($gui{menuitemAttachHD}, $gui{menuitemAttachDVD});
            &sensitive_on($gui{menuitemAttachFloppy});
        }
        else {
            &sensitive_on($gui{menuitemAttachHD});
            &sensitive_off($gui{menuitemAttachFloppy}, $gui{menuitemAttachDVD});
        }
    }
    else {
        &sensitive_on($gui{buttonEditStorRemoveAttach});
        &sensitive_off(@{$sensgrp{ctr_sel_off}});
    }
}

sub tgr_editstor_ctrtype_fill() {
    my ($ctype, $IStorCtr) = @_;
    $gui{comboboxEditStorCtrType}->signal_handler_block($signal{stortype});
    &clear_liststore($gui{liststoreEditStorCtrType});
    my $cvariant = vboxService->IStorageController_getControllerType($IStorCtr);

    if ($ctype eq 'IDE Controller') {
        &append_liststore($gui{liststoreEditStorCtrType}, (0 => 'PIIX3'));
        &append_liststore($gui{liststoreEditStorCtrType}, (0 => 'PIIX4'));
        &append_liststore($gui{liststoreEditStorCtrType}, (0 => 'ICH6'));
    }
    elsif ($ctype eq 'SATA Controller') {
        &append_liststore($gui{liststoreEditStorCtrType}, (0 => 'IntelAhci'));
    }
    elsif ($ctype eq 'SAS Controller') {
        &append_liststore($gui{liststoreEditStorCtrType}, (0 => 'LsiLogicSas'));
    }
    elsif ($ctype eq 'SCSI Controller') {
        &append_liststore($gui{liststoreEditStorCtrType}, (0 => 'LsiLogic'));
        &append_liststore($gui{liststoreEditStorCtrType}, (0 => 'BusLogic'));
    }
    elsif ($ctype eq 'Floppy Controller') {
        &append_liststore($gui{liststoreEditStorCtrType}, (0 => 'I82078'));
    }

    if ($cvariant eq 'ICH6') { $gui{comboboxEditStorCtrType}->set_active(2); }
    elsif ($cvariant eq 'PIIX4' || $cvariant eq 'BusLogic') { $gui{comboboxEditStorCtrType}->set_active(1); }
    else { $gui{comboboxEditStorCtrType}->set_active(0); }
    $gui{comboboxEditStorCtrType}->signal_handler_unblock($signal{stortype});
}

sub tgr_editstor_ctrtype() {
    my ($widget, $treeview) = @_;
    my ($cname, $cname2, $ctype, $IStorCtr) = &get_selected($treeview);
    vboxService->IStorageController_setControllerType($IStorCtr, &get_selected($widget, 0));
    &tgr_editstor_tree_fill($EIMachineM);
}

sub tgr_editaudio() {
    my ($widget, $focus) = @_;
    my $IAudioAdapter = vboxService->IMachine_getAudioAdapter($EIMachineM);
    vboxService->IAudioAdapter_setEnabled($IAudioAdapter, $widget->get_active());
}

sub tgr_editaudio_ctr() {
    my ($widget) = @_;
    my $IAudioAdapter = vboxService->IMachine_getAudioAdapter($EIMachineM);
    my $controller = &get_selected($widget, 1);
    vboxService->IAudioAdapter_setAudioController($IAudioAdapter, $controller);
}

sub tgr_editaudio_driver() {
    my ($widget) = @_;
    my $IAudioAdapter = vboxService->IMachine_getAudioAdapter($EIMachineM);
    my $driver = &get_selected($widget, 1);
    vboxService->IAudioAdapter_setAudioDriver($IAudioAdapter, $driver);
}

sub tgr_editnet_adapter() {
    my ($widget, $focus, $checkbutton) = @_;
    my ($INetworkAdapter, $num) = &callout_getnetadapter($checkbutton);
    vboxService->INetworkAdapter_setEnabled($INetworkAdapter, $widget->get_active());
}

sub tgr_editnet_cable() {
    my ($widget, $focus, $checkbutton) = @_;
    my ($INetworkAdapter, $num) = &callout_getnetadapter($checkbutton);
    vboxService->INetworkAdapter_setCableConnected($INetworkAdapter, $widget->get_active());
}

sub tgr_editnet_mac() {
    my ($widget, $focus, $checkbutton) = @_;
    my ($INetworkAdapter, $num) = &callout_getnetadapter($checkbutton);
    my $mac = $widget->get_text();
    vboxService->INetworkAdapter_setMACAddress($INetworkAdapter, $mac) if (length($mac) == 12);
}

sub tgr_editnet_adapter_type() {
    my ($widget, $checkbutton) = @_;
    my ($INetworkAdapter, $num) = &callout_getnetadapter($checkbutton);
    my $type = &get_selected($widget, 1);
    vboxService->INetworkAdapter_setAdapterType($INetworkAdapter, $type);
}

sub tgr_editnet_attach() {
    my ($widget, $checkbutton) = @_;
    my ($INetworkAdapter, $num) = &callout_getnetadapter($checkbutton);
    $gui{'comboboxEditNetName' . $num}->signal_handler_block($signal{'netname' . $num});
    my $IHost = vboxService->IVirtualBox_getHost($gui{websn});
    my $AttachType = &get_selected($widget, 1);
    &clear_liststore($gui{'liststoreEditNetName' . $num});

    if ($AttachType eq 'Bridged') {
        # Set current bridged network name for NICs 1-4, depending on attachment type
        my @IHostNetworkInterface = vboxService->IHost_getNetworkInterfaces($IHost);
        foreach my $interface (@IHostNetworkInterface) {
            if (vboxService->IHostNetworkInterface_getInterfaceType($interface) eq 'Bridged') {
                my $name = vboxService->IHostNetworkInterface_getName($interface);
                my $iter = &append_liststore($gui{'liststoreEditNetName' . $num}, (0 => $name));
            }
        }
        vboxService->INetworkAdapter_attachToBridgedInterface($INetworkAdapter);
        &hide_off($gui{'labelEditNetName' . $num}, $gui{'comboboxEditNetName' . $num});
        &hide_on($gui{'labelEditNetNameInt' . $num}, $gui{'entryEditNetNameInt' . $num});
    }
    elsif ($AttachType eq 'HostOnly') {
        # Set current hostonly network name for NICs 1-4, depending on attachment type
        my @IHostNetworkInterface = vboxService->IHost_getNetworkInterfaces($IHost);
        foreach my $interface (@IHostNetworkInterface) {
            if (vboxService->IHostNetworkInterface_getInterfaceType($interface) eq 'HostOnly') {
                my $name = vboxService->IHostNetworkInterface_getName($interface);
                my $iter = &append_liststore($gui{'liststoreEditNetName' . $num}, (0 => $name));
            }
        }
        vboxService->INetworkAdapter_attachToHostOnlyInterface($INetworkAdapter);
        &hide_off($gui{'labelEditNetName' . $num}, $gui{'comboboxEditNetName' . $num});
        &hide_on($gui{'labelEditNetNameInt' . $num}, $gui{'entryEditNetNameInt' . $num});
    }
    elsif ($AttachType eq 'Internal') {
        my $name = vboxService->INetworkAdapter_getInternalNetwork($INetworkAdapter);
        $gui{'entryEditNetNameInt' . $num}->set_text($name);
        vboxService->INetworkAdapter_attachToInternalNetwork($INetworkAdapter);
        &hide_on($gui{'labelEditNetName' . $num}, $gui{'comboboxEditNetName' . $num});
        &hide_off($gui{'labelEditNetNameInt' . $num}, $gui{'entryEditNetNameInt' . $num});
    }
    elsif ($AttachType eq 'NAT') {
        vboxService->INetworkAdapter_attachToNAT($INetworkAdapter);
        &hide_on($gui{'labelEditNetName' . $num}, $gui{'comboboxEditNetName' . $num},
                 $gui{'labelEditNetNameInt' . $num}, $gui{'entryEditNetNameInt' . $num});
    }
    else {
        # Assume detach as we either don't support it yet, or it really is detach
        vboxService->INetworkAdapter_detach($INetworkAdapter);
        &hide_on($gui{'labelEditNetName' . $num}, $gui{'comboboxEditNetName' . $num},
                 $gui{'labelEditNetNameInt' . $num}, $gui{'entryEditNetNameInt' . $num});
    }

    $gui{'comboboxEditNetName' . $num}->signal_handler_unblock($signal{'netname' . $num});
    $gui{'comboboxEditNetName' . $num}->set_active(0); # This will call tgr_editnet_name()
}

sub tgr_editnet_name() {
    my ($widget, $checkbutton) = @_;
    my ($INetworkAdapter, $num) = &callout_getnetadapter($checkbutton);
    my $AttachType = &get_selected($gui{'comboboxEditNetAttach' . $num}, 1);
    if ($AttachType eq 'Bridged' | $AttachType eq 'HostOnly') {
        vboxService->INetworkAdapter_setHostInterface($INetworkAdapter, $gui{'comboboxEditNetName' . $num}->get_active_text());
    }
}

sub tgr_editnet_name_internal() {
    my ($widget, $focus, $checkbutton) = @_;
    my ($INetworkAdapter, $num) = &callout_getnetadapter($checkbutton);
    my $name = $gui{'entryEditNetNameInt' . $num}->get_text();
    vboxService->INetworkAdapter_setInternalNetwork($INetworkAdapter, $name) if ($name);
}

sub callout_getnetadapter() {
    my ($checkbutton) = @_;
    my $slot = 0;
    if ($checkbutton eq $gui{checkbuttonEditNetEnable1}) { $slot = 0; }
    elsif ($checkbutton eq $gui{checkbuttonEditNetEnable2}) { $slot = 1; }
    elsif ($checkbutton eq $gui{checkbuttonEditNetEnable3}) { $slot = 2; }
    elsif ($checkbutton eq $gui{checkbuttonEditNetEnable4}) { $slot = 3; }
    my $INetworkAdapter = vboxService->IMachine_getNetworkAdapter($EIMachineM, $slot);
    return $INetworkAdapter, $slot+1;
}

sub tgr_editusb() {
    my ($widget, $focus) = @_;
    my $IUSBController = vboxService->IMachine_getUSBController($EIMachineM);
    vboxService->IUSBController_setEnabled($IUSBController, $widget->get_active());
}

sub tgr_editusb_ehci() {
    my ($widget, $focus) = @_;
    my $IUSBController = vboxService->IMachine_getUSBController($EIMachineM);
    vboxService->IUSBController_setEnabledEhci($IUSBController, $widget->get_active());
}

sub tgr_vmm_selected() {
    my ($widget, $liststore) = @_;
    my ($name, $IMedium, $gnames, $asize, $vsize) = &get_selected($widget);

    if ($gnames eq '<None>') {
        &sensitive_on($gui{toolbuttonVMMRemove});
        &sensitive_off($gui{toolbuttonVMMRelease});
    }
    else {
        &sensitive_off($gui{toolbuttonVMMRemove});
        &sensitive_on($gui{toolbuttonVMMRelease});
    }
}

sub tgr_vmm_tabchanged() {
    my ($widget, $focus, $page) = @_;
    &sensitive_off($gui{toolbuttonVMMRemove}, $gui{toolbuttonVMMRelease});
    # We need to re-evaluate selection on tab change
    if ($page == 0) {
        $gui{buttonVMMAdd}->set_label('_Add Hard Disk');
        $gui{labelVMMLocation}->set_text_with_mnemonic('HD _Location');
        my $selected = $gui{treeviewVMMHD}->get_selection->get_selected();
        &tgr_vmm_selected($gui{treeviewVMMHD}, $gui{liststoreVMMHD}) if ($selected);
    }
    elsif ($page == 1) {
        $gui{buttonVMMAdd}->set_label('_Add CD/DVD');
        $gui{labelVMMLocation}->set_text_with_mnemonic('CD/DVD _Location');
        my $selected = $gui{treeviewVMMDVD}->get_selection->get_selected();
        &tgr_vmm_selected($gui{treeviewVMMDVD}, $gui{liststoreVMMDVD}) if ($selected);
    }
    else {
        $gui{buttonVMMAdd}->set_label('_Add Floppy');
        $gui{labelVMMLocation}->set_text_with_mnemonic('Floppy _Location');
        my $selected = $gui{treeviewVMMFloppy}->get_selection->get_selected();
        &tgr_vmm_selected($gui{treeviewVMMFloppy}, $gui{liststoreVMMFloppy}) if ($selected);
    }
}

sub tgr_vmm_release() {
    my ($widget, $notebook) = @_;
    my $page = $notebook->get_current_page();
    my ($name, $IMedium);
    my $warn = 0;
    &busy_window($gui{dialogVMM});

    if ($page == 0) { ($name, $IMedium) = &get_selected($gui{treeviewVMMHD}); }
    elsif ($page == 1) { ($name, $IMedium) = &get_selected($gui{treeviewVMMDVD}); }
    else { ($name, $IMedium) = &get_selected($gui{treeviewVMMFloppy}); }

    if (vboxService->IMedium_getChildren($IMedium)) {
        &show_warn_dialog('Cannot Release Medium',
                          'This medium appears to have snapshots associated with it. Due to current' .
                          ' limitations in RemoteBox, media with snapshots cannot yet be released');
    }
    else {
        my @guuids = vboxService->IMedium_getMachineIds($IMedium);

        foreach my $id (@guuids) {
            my $IMachine = vboxService->IVirtualBox_getMachine($gui{websn}, $id);
            my @IMediumAttachment = vboxService->IMachine_getMediumAttachments($IMachine);

            foreach my $attach (@IMediumAttachment) {

                if ($$attach{medium} eq $IMedium) {
                    my ($ISession, $IMachineM) = &get_mutable_session($IMachine);

                    if ($IMachineM) {
                        vboxService->IMachine_detachDevice($IMachineM, $$attach{controller}, $$attach{port}, $$attach{device});
                        vboxService->IMachine_saveSettings($IMachineM);
                        vboxService->ISession_close($ISession);
                    }
                    else { $warn = 1; }
                }
            }
        }
    }

    if ($warn == 1) {
        &show_warn_dialog('Medium In Use',
                          'The selected medium cannot be released from all its guests. Some of the' .
                          ' medium\'s guests are in a state which does not permit releasing');
    }

    &set_status_msg("Storage $name released (or partially released)");

    if ($page == 0) { &vmm_mediahd_fill(); }
    elsif ($page == 1) { &vmm_mediadvd_fill(); }
    else { &vmm_mediafloppy_fill(); }

    &unbusy_window($gui{dialogVMM});
}

sub tgr_vmm_rem() {
    my ($widget, $notebook) = @_;
    my $page = $notebook->get_current_page();
    &busy_window($gui{dialogVMM});

    if ($page == 0) {
        my ($name, $IMedium) = &get_selected($gui{treeviewVMMHD});
        if (vboxService->IMedium_getChildren($IMedium)) {
            &show_warn_dialog('Cannot Remove Medium',
                              'This medium appears to have snapshots associated with it. Due to current' .
                              ' limitations in RemoteBox, media with snapshots cannot yet be removed');
        }
        else {
            &busy_window($gui{dialogVMM});
            my $response = $gui{dialogRemoveDelete}->run;
            $gui{dialogRemoveDelete}->hide;

            if ($response eq '1') { # Delete
                &set_status_msg("Deleting hard disk $name...");
                #vboxService->IMedium_close($IMedium);
                my $IProgress = vboxService->IMedium_deleteStorage($IMedium);

                do {
                    &set_status_msg('Deleting hard disk... ' . vboxService->IProgress_getPercent($IProgress) . '% complete.');
                    sleep 1; # Poll only every 1 second
                } until (vboxService->IProgress_getCompleted($IProgress) eq 'true');

                &set_status_msg("Hard disk $name deleted.");
                &vmm_mediahd_fill();
            }
            elsif ($response eq '2') { # Remove
                vboxService->IMedium_close($IMedium);
                &set_status_msg("Storage $name removed.");
                &vmm_mediahd_fill();
            }
            &unbusy_window($gui{dialogVMM});
        }
    }
    elsif ($page == 1) {
        my ($name, $IMedium) = &get_selected($gui{treeviewVMMDVD});
        vboxService->IMedium_close($IMedium);
        &set_status_msg("Storage $name removed.");
        &vmm_mediadvd_fill();
    }
    else {
        my ($name, $IMedium) = &get_selected($gui{treeviewVMMFloppy});
        vboxService->IMedium_close($IMedium);
        &set_status_msg("Storage $name removed.");
        &vmm_mediafloppy_fill();
    }
    &unbusy_window($gui{dialogVMM});
}

sub tgr_vmm_add() {
    my ($widget, $pathentry) = @_;
    &busy_window($gui{dialogVMM});
    my $location = $pathentry->get_text();
    $pathentry->set_text('');
    my $page = $gui{notebookVMM}->get_current_page();

    if ($location) {
        if ($page == 0) {
            vboxService->IVirtualBox_openHardDisk($gui{websn}, $location, 'ReadWrite', 0, '', 0, '');
            &vmm_mediahd_fill();
        }
        elsif ($page == 1) {
            vboxService->IVirtualBox_openDVDImage($gui{websn}, $location, undef);
            &vmm_mediadvd_fill();
        }
        else {
            vboxService->IVirtualBox_openFloppyImage($gui{websn}, $location, undef);
            &vmm_mediafloppy_fill();
        }
    }

    &unbusy_window($gui{dialogVMM});
}

sub tgr_name_validate() {
    my ($entry, $char, $len, $pos) = @_;
    $char =~ s/[\?\/\;\*\\\<\>\|\.]//; # Strip these chars
    return $char, $pos;
}

sub tgr_mac_validate() {
    my ($entry, $char, $len, $pos) = @_;
    $char =~ s/[^A-F0-9a-f]//; # Strip everything but these chars
    return $char, $pos;
}

sub tgr_port_validate() {
    my ($entry, $char, $len, $pos) = @_;
    $char =~ s/[^0-9,-]//; # Strip everything but these chars
    return $char, $pos;
}

sub tgr_timeout_validate() {
    my ($entry, $char, $len, $pos) = @_;
    $char =~ s/[^0-9]//; # Strip everything but these chars
    return $char, $pos;
}